function [y,lud] = backslash(A,b,x,lud,right)
% Reworked backslash operator, so that a function handle A and saving the
% LU data is possible. Uses a correction in case A forms a sparse matrix
% (based on code from Roel Van Beeumen)
%
% Inputs:
%   A   - Matrix or matrix-valued function
%   b   - Right hand side
%   x   - If A is a function handle, the point to evaluate A. In case A is a matrix, this input is ignored.
%   lud - LU information
%       lud.LU  - cell containing LU data, either L,U or L,U,P,Q,R in case
%       of a sparse matrix
%       lud.stat - current status of lu data [0], either 
%           0 - no lu data given and no lu data wanted,
%           1 - no lu data given but lu data wanted (i.e. get LU),
%           2 - sparse lu data given
%           3 - full lu data given
%   right - If true, do forward slash instead, i.e. b/A


% Rework the inputs
if nargin < 3 || isempty(x), x = 0; end
if nargin < 4 || isempty(lud)
    lud = struct;  % Set lud to empty structure
end
if ~isfield(lud,'stat')
    lud.stat = 0;  % No LU data given, no LU data wanted is default
end
if nargin < 5, right = 0; end

if lud.stat < 2 % Case of no LU data given
    % Evaluate A if it is a matrix handle, and store the evaluation in lud
    if isa(A,'function_handle')
        lud.A = A(x);
    else
        lud.A = A;
    end
    % Get LU if necessary
    if lud.stat == 1
        if issparse(lud.A)
            lud.stat = 2; % Go to status 2, sparse matrix LU data is stored
            lud.LU = cell(1,5);
            [lud.LU{:}] = lu(lud.A);
        else
            lud.stat = 3; % Status 3, full matrix LU data is stroed
            lud.LU = cell(1,2);
            [lud.LU{:}] = lu(lud.A);
        end
    end
end
if ~right % Check if it is really a backslash or forward slash
    % Reshape b in case of multiple right hand sides that have been
    % vectorized
    s = size(b,2);
    b = reshape(b,size(lud.A,1),[]);
    % Get solution to the system
    if lud.stat == 2 % Sparse LU case
        y = lud.LU{4}*(lud.LU{2}\(lud.LU{1}\(lud.LU{3}*(lud.LU{5}\b))));
        % Make sure the result is accurate:
        resid = b - lud.A*y;
        if norm(resid,1) > 1e-14
            err = lud.LU{4}*(lud.LU{2}\(lud.LU{1}\(lud.LU{3}*(lud.LU{5}\resid))));
            y = y + err;
        end
    elseif lud.stat == 3 % Full LU case
        y = lud.LU{2}\(lud.LU{1}\b);
    else % Normal backslash operator
        y = lud.A\b;
    end
    % Reshape to the original second dimension
    y = reshape(y,[],s);
else % Case of forward slash
    s = size(b,1);
    b = reshape(b,[],size(lud.A,2));
    if lud.stat == 2 % Sparse LU case
        y = ((((b*lud.LU{4})/lud.LU{2})/lud.LU{1})*lud.LU{3})/lud.LU{5};
        % Make sure the result is accurate:
        resid = b-y*lud.A;
        if norm(resid,1) > 1e-14
            err = ((((resid*lud.LU{4})/lud.LU{2})/lud.LU{1})*lud.LU{3})/lud.LU{5};
            y = y + err;
        end
    elseif lud.stat == 3
        y = b/lud.LU{1}/lud.LU{2};
    else
        y = b/lud.A;
    end
    y = reshape(y,s,[]);
end
if(lud.stat == 0), lud = rmfield(lud,'A'); end
end