function [e,V,W,info] = eign(A,domain,opts)
% EIGN - Determine the eigenvalues of the non-linear matrix function A
%
% Inputs:
%   - A is a structure containing C,f,p -
%           * C are the coefficient matrices
%           * p is an array containing scalar coefficients for a possible
%           monomial part, so that the polynomial part of P(s) =
%           p(1)*C{1}*s^0 + p(2)*C{2}*s^1 + ...
%           * f is a cell containing the non-linear, non-monomial functions
%     So, A(s) = A.p(1)*A.C{1}*s^0 + ...+ A.p(end)*A.C{length(A.p)-1}*s^(length(A.p)-1) + ...
%     A.f{1}(s) A.C{length(A.p)} + ... + A.f{end}(s) C{end}.
%     We should have: nnz(A.p)+length(A.f) = length(A.C)
%
%   - Sigma is the sample domain of the function, given as a series of
%     points within the interval of interest, or as a lower and upper bound
%     (domain = [a,b]), in which case 1e5 real are chosen from a to b (in case of imaginary a and b, inside a square of the complex plain with lower corner a and upper corner b).
%
%   - opts are the options, given as a structure, that can contain:
%       opts.kind       'taylor','aaa' or 'ratNewton' [aaa]
%       opts.force_t    force transpose instead of conjugate transpose [false]
%       opts.do_left    use the transposed pencil to find the eigenvalues
%                       and vectors
%       opts.twos       if opts.twos is true, also give the left eigenvectors [false]
%       opts.upper_orth only orthogonalize the 'upper part' of the pencil
%       [if twosided: true, if onesided: false]
%       opts.full_orth  orthogonalize fully (!can be very expensive)
%       opts.use_symm   use the symmetricaly structured pencil
%
%       opts.shifts     shifts used during the Krylov routine [0]
%       opts.k          number of eigenvalues to find [10]
%       opts.m          maximum number of iterations of the krylov routine [5*opts.k]
%       opts.v0         starting vector for the (right) Krylov iterations [rand(n,1)+1i*rand(n,1)]
%       opts.v0_s0      if this value is given, the starting vector will be (AA-opts.v0_s0*BB)\opts.v0 []
%       opts.w0         same as opts.v0 but for left space [opts.v0]
%       opts.w0_s0      same as opts.v0_s0 but for the left space [opts.v0_s0]
%       opts.restarts   a vector with the iteration numbers in which to use restarts
%
%       opts.Xi         (if opts.kind = 'Newton') the sample set for the poles
%       opts.d          (if opts.kind = 'Newton') degree of the Newton poly
%       opts.verbose    printstatements and plots related to the computation are given
%       opts.tol_appr   (!only for AAA) tolerance for the approximation of the non-polynomial part [1e-13]
%       opts.norms      the norms of the factor matrices
%       opts.der        (!only for Taylor) derivative coefficients, so that [C{1} ...
%       C{end}]*kron(opts.der,speye(n,n)) = [P{1} ... P{d}], with
%       P{1},...,P{d} the coefficient factors
%
%       opts.pos_im     use Re(lambda) + |Im(lambda)| for tolerance computation [false]
%       opts.scale_res  scale the residue with the norms of the matrices [true]
%
% Outputs:
%   - e - eigenvalues
%   - V - Right eigenvectors
%   - W - Left eigenvectors
%   - info - structure containing the following fields:
%       info.A_,info.B_     Projected linear pencil matrices

% Work out the options
[get_res,kind,shifts,k,m,use_symmetric,trnsp,tol_appr,pos_im,scale_res,upper_orth,full_orth,proj,verbose] = checkopts(opts);

% Work out the inputs
n = size(A.C{1},1);
nC = length(A.C);
if ~isfield(opts,'norms') && scale_res % No norms given and need to scale with norms
    norms = cellfun(@(x) norm(x,1),A.C);
elseif scale_res
    norms = opts.norms;
end
% Handle the domain
if length(domain) == 2
    if isreal(a) && isreal(b)
        domain = linspace(domain(1),domain(2),1e5);
    else
        domain = linspace(real(domain(1)),real(domain(2)),320)' + 1i*linsapce(imag(domain(1)),imag(domain(2)),320);
    end
end
domain = domain(:);

% Monomial part handeling
% TODO: better handling of only constant term (Does not work yet?)
if ~isfield(A,'p') || isempty(A.p) % Case of no monomials given
    A.p = [];
    d_p = 0;
    n_p = 0;
elseif length(A.p) == 1
        cc = A.p;
        A.f = [{@(x) cc*ones(size(x))},A.f];
        A.p = [];
else
    % Add the polynomial part to the function evalutions
    d_p = length(A.p);                 % Degree of polynomial
    A.p = A.p(:);                      % Make sure p is a row vector
    n_p = sum(A.p~=0);                 % Number of terms that are non-zero
    T = spdiags(A.p(A.p~=0),0,nC,n_p); % Weight matrix for monomial part
    % Differentiate between symmetric and non-symmetric
    if ~use_symmetric
        TA_p = spdiags(logical(A.p(1:end-1)),0,n_p,d_p-1);
        TB_p = sparse(n_p,d_p-1,-1,n_p,d_p-1);
        p.M_phi = spdiags(ones(d_p-1,1),1,d_p-2,d_p-1);
        p.N_phi = spdiags(ones(d_p-1,1),0,d_p-2,d_p-1);
    else
        TA_p = sparse(find(A.p(1:2:end)~=0),find(A.p(1:2:end)~=0),logical(A.p(1:2:end)),n_p,ceil(d_p/2));
        TB_p = sparse(find(A.p(2:2:end)~=0),find(A.p(2:2:end)~=0),-logical(A.p(2:2:end)),n_p,ceil(d_p/2));
        
        p.M_phi = spdiags(ones(ceil(d_p/2)+1,1),1,ceil(d_p/2)-1,ceil(d_p/2));
        p.N_phi = spdiags(ones(ceil(d_p/2)+1,1),0,ceil(d_p/2)-1,ceil(d_p/2));
    end
end

% Convert non-linear functions in cell to array-valued functions
f_v = @(x) cell2mat(cellfun(@(fi) feval(fi,x),A.f,'UniformOutput',false));

switch kind
    case 'taylor'
        if d_p > 0
            d = size(opts.der,2)-1;
            T = [[T,sparse(n_p,d-size(T,2))];sparse(length(A.C),d)] +  [sparse(n_p,d);opts.der];
            %             T = opts.der;
            p.M_phi = spdiags([opts.shifts./(1:d).' ones(d,1)],[0,1],d,d+1);
            p.N_phi = spdiags(ones(d,1)./(1:d).',0,d,d+1);
            p.M_psi = zeros(0,1);
            p.N_psi = zeros(0,1);
            p.TA = T*spdiags(ones(d,1),0,d,d-1);
            p.TB = T*[sparse(d,d-1) -flipud(speye(d,1))];
            p.L_phi = 1:d-1+size(TA_p,2);
            p.L_psi = ones(1,d-1+size(TA_p,2));
            lc = [];
            g = @(x) 1;
        elseif use_symmetric
            D = size(opts.der,2);
            d = (size(opts.der,2)-2)/2;
            T = opts.der;
            
            p.M_phi = spdiags([opts.shifts./(2:2:D-1).' ones(d,1)],[0,1],d,d+1);
            p.N_phi = spdiags(ones(d,1)./(2:2:D-1).',0,d,d+1);
            p.M_psi = spdiags([opts.shifts./(3:2:D-1).' ones(d,1)],[0,1],d,d+1);
            p.N_psi = spdiags(ones(d,1)./(3:2:D-1).',0,d,d+1);
            
            p.TA = T*sparse(1:2:D,1:d+1,1:2:D,D,d+1);
            p.TB = T*sparse(2:2:D,1:d+1,-ones(1,d+1),D,d+1);
            
            p.L_phi = 1:d+1;
            p.L_psi = 1:d+1;
            lc = [];
            g = @(x) 1;
        else
            d = size(opts.der,2)-1;
            T = opts.der;
            p.M_phi = spdiags([opts.shifts./(1:d).' ones(d,1)],[0,1],d-1,d);
            p.N_phi = spdiags(ones(d,1)./(1:d).',0,d-1,d);
            p.M_psi = zeros(0,1);
            p.N_psi = zeros(0,1);
            p.TA = T*[[sparse(1,1,1,2,2) sparse(2,d-2)];spdiags(ones(d,1),2,d-1,d)];
            p.TB = T*[sparse(2,1,-1,size(T,2),1)  sparse(d+1,d-2) -flipud(speye(d+1,1))];
            p.L_phi = 1:d;
            p.L_psi = ones(1,d);
            lc = [];
            g = @(x) 1;
        end
    case 'ratNewton'
        % Get degree
        d = opts.d;
        % Get the Newton variables
        [sig,pol,beta] = lejabagby(domain,opts.Xi,domain,d+1,false,1);
        sig = sig(:); xi = pol(:); beta = beta(:);
        if d_p > 2
            % Transformation from coefficient matrices to polynomial factors
            ff = ratnewtoncoeffs2(f_v,sig,xi,beta,nC-n_p);
            T(n_p+1:end,1) = ff(:,1);
            T = [T,[sparse(n_p,d);ff(:,2:end)]];
            % Transformation from polynomial factors to A,B form, placement of
            % A,B in pencil and linear relation matrices
            TA_n = spdiags(ones(d,1),0,d,d-1);
            TB_n = sparse(d,d-1);
            if ~use_symmetric
                p.M_phi = [[p.M_phi,sparse(d_p-2,d-1)];[sparse(1,1,sig(1),d-1,d_p-2) spdiags([[0;sig(2:end-1)],beta(2:end)],[0,1],d-1,d)]];
                p.N_phi = [[p.N_phi,sparse(d_p-2,d-1)];[sparse(1,1,1,d-1,d_p-2) spdiags([[0;ones(d-1,1)],beta(2:end)./xi(1:end-1)],[0,1],d-1,d)]];
                p.M_psi = zeros(0,1);
                p.N_psi = zeros(0,1);
                p.TA = T*[[TA_p,sparse(n_p,d-1)];[sparse(d,size(TA_p,2)),TA_n]];
                p.TB = T*[[TB_p,sparse(n_p,d-1)];[sparse(d,size(TB_p,2)),TB_n]];
                p.L_phi = 1:d-1+size(TA_p,2);
                p.L_psi = ones(1,d-1+size(TA_p,2));
            else
                p.M_psi = p.M_phi;
                p.N_psi = p.N_phi;
                p.TA = [[TA_p,sparse(n_p,2*d)];[sparse(d,n_p),kron([1,1],0.5*TA_n)]];
                p.TB = [[TB_p,sparse(n_p,2*d)];[sparse(d,n_p),kron([1,1],0.5*TB_n)]];
                p.L_phi = [1:ceil(d_p/2),ceil(d_p/2)+1:d+ceil(d_p/2),ones(1,d)];
                p.L_psi = [1:ceil(d_p/2),ones(1,d),ceil(d_p/2)+1:d+ceil(d_p/2)];
            end
            g = @(x) 1;
            lc = [];
        elseif d_p > 0
            d = d+1;
            % Transformation from coefficient matrices to polynomial factors
            ff = ratnewtoncoeffs2(f_v,sig,xi,beta,nC-n_p);
            T(n_p+1:end,1) = ff(:,1);
            T = [T,[sparse(n_p,d);ratnewtoncoeffs2(f_v,sig,xi,beta,nC-n_p)]];
            % Transformation from polynomial factors to A,B form, placement of
            % A,B in pencil and linear relation matrices
            TA_n = spdiags(ones(d,1),0,d,d);
            TB_n = sparse(d,d);
            % TODO: Trow out the second 1-term, caused by Newton approximation
            p.M_phi = [[p.M_phi,sparse(d_p-2,d)];[-1,sparse(1,d_p-2),speye(1,d)];[sparse(d-1,d_p-1) spdiags([sig(1:end-1),beta(2:end)],[0,1],d-1,d)]];
            p.N_phi = [[p.N_phi,sparse(d_p-2,d)];sparse(1,d+d_p-1);[sparse(d-1,d_p-1) spdiags([ones(d-1,1),beta(2:end)./xi(1:end-1)],[0,1],d-1,d)]];
            if ~use_symmetric
                p.M_psi = zeros(0,1);
                p.N_psi = zeros(0,1);
                p.TA = T*[[TA_p,sparse(n_p,d)];[sparse(d,size(TA_p,2)),TA_n]];
                p.TB = T*[[TB_p,sparse(n_p,d)];[sparse(d,size(TB_p,2)),TB_n]];
                p.L_phi = 1:d+size(TA_p,2);
                p.L_psi = ones(1,d+size(TA_p,2));
            else
                p.M_psi = p.M_phi;
                p.N_psi = p.N_phi;
                p.TA = T*[[TA_p,sparse(n_p,2*d)];[sparse(d,size(TA_p,2)),kron([1,1],0.5*TA_n)]];
                p.TB = T*[[TB_p,sparse(n_p,2*d)];[sparse(d,size(TB_p,2)),kron([1,1],0.5*TB_n)]];
                
                p.L_phi = [1:ceil(d_p/2),ceil(d_p/2)+1:d+ceil(d_p/2),ones(1,d)];
                p.L_psi = [1:ceil(d_p/2),ones(1,d),ceil(d_p/2)+1:d+ceil(d_p/2)];
            end
            g = @(x) 1;
            lc = [];
        else
            T = ratnewtoncoeffs2(f_v,sig,xi,beta,nC);
            if ~use_symmetric
                T = ratnewtoncoeffs2(f_v,sig,xi,beta,nC);
                TA = spdiags(ones(d,1),0,d+1,d);
                TB = spdiags(1./xi(1:end-1),0,d+1,d);
                TA(end,end) = -sig(end-1)/beta(end);
                TB(end,end) = -1/beta(end);
                
                p.M_phi = spdiags([sig(1:end-2),beta(2:end-1)],[0,1],d-1,d);
                p.N_phi = spdiags([ones(d-1,1),beta(2:end-1)./xi(2:end-1)],[0,1],d-1,d);
                p.M_psi = zeros(0,1);
                p.N_psi = zeros(0,1);
                p.TA = T*TA;
                p.TB = T*TB;
                p.L_phi = 1:d;
                p.L_psi = ones(1,d);
                g = @(x) 1;
                lc = [];
            else
                di = (d-1)/2;
                TA = sparse(1:2:d+1,1:di+1,ones(1,di+1).',d+1,di+1) + sparse(2:2:d+1,1:di+1,-(sig(1:2:end-1).')./(beta(2:2:end).'),d+1,di+1);
                TB = sparse(1:2:d+1,1:di+1,(ones(1,di+1))./xi(1:2:end-1).',d+1,di+1) + sparse(2:2:d+1,1:di+1,-ones(1,di+1)./(beta(2:2:end).'),d+1,di+1);
                
                p.M_phi = spdiags([sig(1:2:end-2),beta(2:2:end-2)],[0,1],di,di+1);
                p.N_phi = spdiags([ones(di,1),beta(2:2:end-2)./xi(2:2:end-1)],[0,1],di,di+1);
                p.M_psi = spdiags([sig(2:2:end-1),beta(3:2:end-1)],[0,1],di,di+1);
                p.N_psi = spdiags([ones(di,1),beta(3:2:end-1)./xi(3:2:end-1)],[0,1],di,di+1);
                
                p.TA = T*TA;
                p.TB = T*TB;
                
                p.L_phi = 1:di+1;
                p.L_psi = 1:di+1;
                g = @(x) 1;
                lc = [];
            end
        end
        info.pol = pol;
    case 'aaa'
        % Estimate the non-polynomial part of the functions
        [~, pol, ~, ~, zj, fj, wj] = aaa_sv(f_v, domain(:),'tol',tol_appr,'mmax',500);
        
        d = length(zj);
        %         if d >= nC-d_p % Check if the degree found is lower than the number of C-terms
        T = [T,[sparse(n_p,d);fj.']];
        %         else
        % TODO: ADD THIS FUNCTIONALITY
        %             p.T = [p.T,[sparse(d_p,nC);speye(nC,nC)]];
        %         end
        TA_n = spdiags(wj,0,d,d);
        TB_n = sparse(d,d);
        p.M_phi = [[p.M_phi,sparse(d_p-2,d)];[1,sparse(1,d_p-2),-wj(1:end).'];[sparse(d-1,d_p-1),spdiags([zj(1:end-1),-zj(2:end)],[0,1],d-1,d)]];
        p.N_phi = [[p.N_phi,sparse(d_p-2,d)];sparse(1,d+d_p-1);[sparse(d-1,d_p-1),spdiags([ones(d-1,1),-ones(d-1,1)],[0,1],d-1,d)]];
        
        p.M_psi = zeros(0,1);
        p.N_psi = zeros(0,1);
        p.TA = T*[[TA_p,sparse(n_p,d)];[sparse(d,size(TA_p,2)),TA_n]];
        p.TB = T*[[TB_p,sparse(n_p,d)];[sparse(d,size(TB_p,2)),TB_n]];
        p.L_phi = 1:d+size(TA_p,2);
        p.L_psi = ones(1,d+size(TA_p,2));
        
        % TODO: twos-sided AAA?
        g = @(x) 1;
        lc = [];
        info.pol = pol;
end
if d_p > 0 % Add the polynomial part to the function evalutions
    powers = 0:1:d_p-1;
    powers(A.p == 0) = [];
    f_v = @(x) [bsxfun(@times,x.^(powers),A.p(A.p~=0).') f_v(x)];
end

%%%%%%%%%% TO DO: ADD IN CASE d < n_nonlin
% CC = cell(1,size(p.T,2));
% for it_t = 1:size(p.T,2)
%     for it_c = 1:size(p.T,1)
%        CC{it_t} = A.C{it_c}*p.T(it_c,it_t);
%     end
% end
% p.T = eye(size(p.T,2),size(p.T,2));
%%%%%%%%%%

% Construct the linear pencil
L = lpencil(A.C,@(x) g(x)*f_v(x),p);

% Set Krylov options
opts_Krylov = struct('k',k,'m',m,'sig',shifts,'save_LU',true,'getres',get_res);
opts_Krylov.resfun = @(lambda,V,H) residual( lambda,V.Q*(reshape(V.u(:,1,:),size(V.u,1),[])*H));
% Set starting vector
if isfield(opts,'v0')
    v0 = opts.v0;
else
    v0 = rand(n,1)+1i*rand(n,1);
end
% Do a first solve, if needed
if isfield(opts,'v0_s0')
    vv0 = L.bs(v0,opts.v0_s0);
else
    vv0 = L.set(v0,lc);
end
nb = size(v0,2);
% Set orthogonalization function
if full_orth
    opts_Krylov.orth_fun = @(v,V) L.orth(v,nb,0);
elseif upper_orth
    opts_Krylov.orth_fun = @(v,V) L.orth(v,nb,1:size(p.M_phi,2));
else
    opts_Krylov.orth_fun = @(v,V) L.orth(v,nb);
end
% opts_Krylov.orth_fun = @(v,V) L.orth(v,nb);
% Set restarts if needed, and then also set multfun
if isfield(opts,'restarts')
    opts_Krylov.restarts = opts.restarts;
    opts_Krylov.mult_fun = @(v,A) multfun_m1(v,A);
    opts_Krylov.subr_fun = @(V) V;
end
% Set filter if needed, and then also set multfun
if isfield(opts,'filter')
    opts_Krylov.filter = opts.filter;
    opts_Krylov.mult_fun = @(v,A) multfun(v,A);
end
% Set concatenation function
opts_Krylov.conca_fun = @(v,V) v;
% Set verbose
opts_Krylov.verbose = verbose;

if proj ~= 1
    [ V,info ] = rkrylov(@(v,s,lud) L.bsB(v,s,nb,lud),vv0,opts_Krylov );
    if proj == 0
        W = V;
        e = 0;
    end
end
if proj ~= 0
    % Transpose the pencil
    L.transpose(trnsp);
    % Set the strating vector for the left side
    if isfield(opts,'w0')
        w0 = opts.w0;
    elseif isfield(opts,'v0')
        w0 = opts.v0;
    else
        w0 = rand(n,1)+1i*rand(n,1);
    end
    if isfield(opts,'w0_s0')
        ww0 = L.bs(w0,opts.w0_s0);
    else
        ww0 = L.set(w0,lc);
    end
    opts_Krylov.resfun = @(lambda,V,H) residual(lambda,V.Q*(reshape(V.u(:,1,:),size(V.u,1),[])*H),trnsp);
    [ W,info ] = rkrylov(@(v,s,lud) L.bsB(v,s,nb,lud),ww0,opts_Krylov );
    if proj == 1
        V = W;
        e = 0;
    end
    % Retranspose the pencil
    L.transpose(trnsp);
end
% Add the info about the poles
info.pol = pol;

% Project the linear pencil
if proj ~= 1
    [info.A_,info.B_] = L.project(V,W,trnsp);
end
V = V.Q*reshape(V.u(:,1,:),size(V.u,1),[]);
W = W.Q*reshape(W.u(:,1,:),size(W.u,1),[]);

% Two-sided projection:
if proj == 2
    info.Res = zeros(size(info.A_));
    info.Res_left = zeros(size(info.A_));
    info.Lambda = zeros(size(info.B_));
    for itt = 1:size(info.A_,1)
        [Vi,E,Wi] = eig(info.A_(1:itt,1:itt),info.B_(1:itt,1:itt));
        info.Res(1:itt,itt) = residual(diag(E),V(:,1:itt)*Vi);
        if trnsp == 1
            info.Res_left(1:itt,itt) = residual(diag(E),conj(W(:,1:itt))*Wi,trnsp);
        else
            info.Res_left(1:itt,itt) = residual(diag(E),W(:,1:itt)*Wi,trnsp);
        end
        info.Lambda(1:itt,itt) = diag(E);
    end
    e = diag(E);
end

    function v = multfun(v,A)
        [rr,dd,kk] = size(v.u);
        uu = reshape(v.u(:,:,:),[],kk);
        uu = uu*A;
        uu = reshape(uu,rr,dd*size(A,2));
        
        [VV,SS,~] = svd(uu,'econ');
        kk = sum(diag(SS) > 1e-15);
        v.Q = v.Q*VV(:,1:kk);
        uu = VV(:,1:kk)'*uu;
        
        v.u = reshape(uu,[],dd,size(A,2));
    end

    function r = residual(lamm,v,trnsp)
        % Expensive residual computation based on the error of the original
        % non-linear function
        if nargin < 3, trnsp = 0; end
        if trnsp > 1, lamm = conj(lamm); end
        lam = lamm;
        if pos_im
            for i = 1:length(lamm)
                lam(i) = real(lamm(i)) + 1i*(sign(imag(lamm(i)))*imag(lamm(i)));
            end
        end
        fi = f_v(lam(:));
        if scale_res 
            den = abs(fi)*norms(:);
        else
            den = abs(fi);
        end
        for it = 1:size(v,2)
            v(:,it) = v(:,it)/norm(v(:,it));
        end
        r = zeros(length(lam),1);
        for it = 1:length(lam)
            if trnsp == 0
                vi = zeros(n,1);
            else
                vi = zeros(1,n);
            end
            for it_in = 1:nC
                switch trnsp
                    case 0
                        vi = vi + A.C{it_in}*v(:,it)*fi(it,it_in);
                    case 1
                        vi = vi + v(:,it).'*A.C{it_in}*fi(it,it_in);
                    case 2
                        vi = vi + v(:,it)'*A.C{it_in}*fi(it,it_in);
                end
            end
            if scale_res
                r(it) = norm(vi)/den(it);
            else
                r(it) = norm(vi);
            end
        end
    end

    function [get_res,kind,shifts,k,m,use_symmetric,trnsp,tol_appr,pos_im,scale_res,upper_orth,full_orth,proj,verbose] = checkopts(opts)
        % Options for the (rational) polynomial approximation
        %
        % Set kind of approximation
        if isfield(opts,'kind')
            kind = opts.kind;
        else
            kind = 'aaa';
        end
        % Tolerance for the approximation (for AAA only)
        if isfield(opts,'tol_appr')
            tol_appr = opts.tol_appr;
        else
            tol_appr = 1e-13;
        end
        
        % Options for the transpose/two-sided
        %
        % Use the transpose instead of the conjugate transpose
        if isfield(opts,'force_t')
            if(opts.force_t)
                trnsp = 1;
            else
                trnsp = 2;
            end
        else
            trnsp = 2;
        end
        % Do two-sided Krylov
        if isfield(opts,'twos')
            do_twosided = opts.twos;
        else
            do_twosided = false;
        end
        % Set use of symmetric pencil
        if isfield(opts,'use_symm')
            use_symmetric = opts.use_symm;
        else % Use symmetric pencil in two-sided case, except for aaa
            use_symmetric = (do_twosided && ~strcmp(kind,'aaa'));
        end
        % Get the left eigenvectors only
        if isfield(opts,'do_left')
            do_left = opts.do_left;
        else
            do_left = false;
        end
        
        if do_twosided
            proj = 2;
        elseif do_left
            proj = 1;
        else
            proj = 0;
        end
        
        % Options for the orthogonalziation
        %
        % Only orthogonalize the upper part
        if isfield(opts,'upper_orth')
            upper_orth = opts.upper_orth;
        elseif use_symmetric
            upper_orth = true;
        else
            upper_orth = false;
        end
        % Do full orthogonalization
        if isfield(opts,'full_orth')
            full_orth = opts.full_orth;
        else
            full_orth = false;
        end
        
        % Options for the Krylov method
        %
        % Shifts
        if isfield(opts,'shifts')
            shifts = opts.shifts;
        else
            shifts = 0;
        end
        % Number of eigenvalues
        if isfield(opts,'k')
            k = opts.k;
        else
            k = 10;
        end
        % Number of iterations
        if isfield(opts,'m')
            m = opts.m;
        else
            m = k*5;
        end
        % Verbose option
        if isfield(opts,'verbose')
            verbose = opts.verbose;
        else
            verbose = false;
        end
        if isfield(opts,'get_res')
            get_res = opts.get_res;
        else
            get_res = true;
        end
        
        % Options for the residue calcualtion
        %
        % Flip negative imaginary parts of eigenvalues to positive values for the
        % computation of the residue
        if isfield(opts,'pos_im')
            pos_im = opts.pos_im;
        else
            pos_im = false;
        end
        % Scale the residues with the norm of the matrices
        if isfield(opts,'scale_res')
            scale_res = opts.scale_res;
        else
            scale_res = true;
        end
    end
end