function model = car()
% Setup the car model with porous seats. Model from Dr. Axel van de Walle
% (KU Leuven)using a mesh from Siemens tutorial, and using poro-elastic material 
% properties from A LLARD , J. & N OUREDDINE , A. (2009) 
% Propagation of sound in porous media: modelling sound absorbing
% materials, 2nd edn. John Wiley & Sons. 
% and C HAZOT , J.-D., N ENNIG , B. & P ERREY -D EBAIN , E. (2013) 
% Performances of the partition of unity finite element
% method for the analysis of two-dimensional interior sound fields with 
% absorbing materials. Journal of Sound and Vibration, 332, 1918–1929.

% Load system matrices
% --------------------
car = load('car.mat');

% Constants
% ---------
% Material properties of air
P0 = 101325;                                                        % Ambient pressure.
gamma=1.4;                                                          % Heat capacity ratio of air, see ideal gas law. gamma = Cp/Cv.
eta = 1.839e-5;                                                     % Dynamic viscosity of air. Value from Fireflex supplier.
rho0=1.213;                                                         % Density of air. Value from Fireflex supplier.
R_gas = 287.031;                                                    % Specific gas constant of air.
T_absolute=P0/(rho0*R_gas);                                         % Absolute tempreature, through ideal gas law
Cp = 4168.8*(0.249679-7.55179e-5*T_absolute+1.69194e-7 ...
     *T_absolute^2-6.46128e-11*T_absolute^3);                       % Specific heat capacity of air.
kappa = 2.624e-02 * ( (T_absolute/300)^(3/2) ...
        * (300+245.4*exp(-27.6/300)) ...
        /(T_absolute+245.4*exp(-27.6/T_absolute)) );                % Thermal conductivity of air.
N_Pr = eta*Cp/kappa;                                                % Prandtl number of air.

% Material properties of the poro-elastic material (from paper)
sigma = 13500;          % Flow resistivity [Ns/m^4] of[kg/(m^3 s)]
phi = 0.98;             % Porosity []
Lambda = 80e-6;         % Viscous characteristic length [m]
Lambda_accent = 160e-6;	% Thermal characteristic length [m]
alfa_inf = 1.7;         % Tortuosity []


% Create the model
% ----------------
model.n = size(car.Ka,1);
model.A.C = {car.Ka,car.Ma,car.Kp,car.Mp};
model.A.p = [1,0,-(2*pi)^2];
model.A.f = {@(x) rho(2*pi*x), @(x) -(2*pi*x).^2.*Bs(2*pi*x)};
model.B = car.F;
model.C = zeros(model.n,1); model.C(10197) = 1;
model.f = 1:300;

    function rho_scaling = rho(freq)
        omega = freq;
        G_J=(1+(4*1i*alfa_inf^2*eta*rho0.*omega)/(sigma^2*Lambda^2*phi^2)).^(1/2); % relaxation function
        alpha = alfa_inf*(ones(size(omega))+(sigma*phi./(1i*omega*rho0*alfa_inf)).*G_J);
        rho_scaling = phi./alpha;
    end

    function B_scaling = Bs(freq)
        omega = freq;
        alpha_accent = ones(size(omega)) + 8*eta./(1i*Lambda_accent^2*N_Pr*omega*rho0) .* sqrt(ones(size(omega)) + 1i*rho0*omega*N_Pr*Lambda_accent^2/(16*eta));
        B_scaling = phi*(gamma - (gamma - 1)./alpha_accent);
    end
end