function L = lpencil(C,f,p)
%LPENCIL - Linear pencil formulation from the non-linear matrix function
%          N(x) = C{1}*fi(1) + C{2}*fi(2) + ... + C{end}fi(end),
%          where fi = f(x), using the polynomial approximation
%          P(x) = (A{1}-x*B{1})*phi_0(x)*psi_0(x) + ... +
%          (A{d}-x*B{d})*phi_d(x)*psi_d(x), where phi_0(x)*psi_0(x) is
%          presumed 1.
%
% Inputs:
%  C    - Matrix coefficients of non-linear matrix function [C_1 ... C_d_C]
%  f    - Matrix functions corresponding to coefficients so that
%  \hat{f}(s) = [f_1(s) f_2(s) ... f_d_C(s)] (s is presumed a row vector),
%  and f = g(s)*\hat{f}(s), where g(s) = 1/\hat{phi_0}(x)*1/\hat{psi_0}(x),
%  i.e. \hat{psi_0}(x) and \hat{phi_0}(x) where the original zeroth
%  polynomial functions, however, since psi_0(x)*phi_0(x) should be 0, f is
%  scaled.
%  p    - Information of the (rational) polynomial approximate:
%             * p.TA and p.TB so that C * kron(p.TA,speye(n,n)) =
%             [A{1},...,A{d}] (and analogues for p.TB)
%             * p.M_phi and p.M_psi so that 
%               (p.M_phi(:,2:end)-x*p.N_phi(:,2:end))\(p.M_phi(:,1)-x*p.N_phi(:,1))
%               = [phi(2),...,phi(d)]', and analogues for p.M_psi
%             * p.L_phi and p.L_psi the positions of A{i} and B{i} in the
%             pencil

% Make the linear pencil
[AA_L,BB_L,d_psi,d_phi,d_C] = initialize(p);
L.id = 'linear pencil';
L.ist = 0; % 0 = not transposed, 1 = transposed, 2 = conjugate transposed

% Pencil functions
L.set = @ set_vector;
% L.ctranspose = @ ctranspose_;
L.transpose = @ transpose_;
% L.ctranspose_mat = @ ctranspose_mat;
% L.transpose_mat = @ transpose_mat;
L.orth = @ orth_u;
L.bs = @bs;
L.bsB = @ bsB;
% L.projectt = @ projectt;
L.project = @ project;
L.large = @large ;
L.ip = @ipp;

    function v = set_vector(v0,lc)
        % Set a vector, given a (n x 1) starting vector, and, optional, a linear
        % combination of the starting vector (only rank-1 starting vectors
        % possible)
        % TODO: expand to non-rank-1 starting vectors
        
        v.Q = v0/norm(v0);
        if nargin < 2 || isempty(lc)
            v.u = norm(v0)*eye(1,size(AA_L,1));     
        else
            v.u = norm(v0)*lc;
        end
    end

    function transpose_(trsp)
        % Transpose the linear pencil
        % TODO: combine with ctranspose
        
        % Check if the linear pencil currently is in transposed form, if
        % so, we will transpose back
        if L.ist == 1
            L.ist = 0;
        else 
            L.ist = 1;
        end
        % Get indices to do the permute on A_L and B_L
        inds = kron(1:d_phi+1,ones(1,(d_psi+1)*d_C)) + ...
            kron(ones(1,d_phi+1),kron(0:d_phi+1:(d_C-1)*(d_phi+1),ones(1,d_psi+1))) + ...
            kron(ones(1,d_C*(d_phi+1)),0:(d_phi+1)*d_C:d_psi*(d_phi+1)*d_C);
        % Get transformation matrices to go from M \kron I_C to M
        Tcl = kron(speye(d_psi),speye(d_C,1)).';
        Tcr = kron(speye(d_psi+1),speye(d_C,1));
        % Get the upper parts of the matrices
        if trsp == 1
        A_L = (AA_L(1:d_phi+1,1:(d_psi+1)*d_C));
        B_L = (BB_L(1:d_phi+1,1:(d_psi+1)*d_C));
        % Reassemble AA_L and BB_L
        AA_L = [ [reshape(A_L(inds),d_psi+1,[]), ...
            (Tcl*AA_L(d_phi+2:end,1:(d_psi+1)*d_C)*Tcr).'] ; ...
            [kron(AA_L(1:d_phi+1,(d_psi+1)*d_C+1:end).',speye(d_C)), ...
            sparse(d_phi*d_C,d_psi)]];
        BB_L = [ [reshape(B_L(inds),d_psi+1,[]), ...
            (Tcl*BB_L(d_phi+2:end,1:(d_psi+1)*d_C)*Tcr).'] ; ...
            [kron(BB_L(1:d_phi+1,(d_psi+1)*d_C+1:end).',speye(d_C)), ...
            sparse(d_phi*d_C,d_psi)]];
        elseif trsp == 2
            AA_L = [ [reshape(A_L(inds),d_psi+1,[]), ...
            (Tcl*AA_L(d_phi+2:end,1:(d_psi+1)*d_C)*Tcr)'] ; ...
            [kron(AA_L(1:d_phi+1,(d_psi+1)*d_C+1:end)',speye(d_C)), ...
            sparse(d_phi*d_C,d_psi)]];
            BB_L = [ [reshape(B_L(inds),d_psi+1,[]), ...
                (Tcl*BB_L(d_phi+2:end,1:(d_psi+1)*d_C)*Tcr)'] ; ...
                [kron(BB_L(1:d_phi+1,(d_psi+1)*d_C+1:end)',speye(d_C)), ...
                sparse(d_phi*d_C,d_psi)]];
        end
        % Conjugate the matrix factors
        if trsp == 1 || trsp == 3
            for it = 1:d_C
                C{it} = C{it}.';
            end
        else
            for it = 1:d_C
                C{it} = C{it}';
            end
        end
        
        % Switch the dimensions
        d_phi_ = d_phi;
        d_phi = d_psi;
        d_psi = d_phi_;
    end

    function [ v,h,normm ] = orth_u(v,nb,range)
        % Orthogonalize the last nb vectors in v.u opposite the rest, where
        % nb is a number of given blocks (in case v is a block vector)
        
        % Size of u
        [r,d,k] = size(v.u); 
        
        if nargin < 3 % Orthogonalize the complete u-vector
            [u_,h,normm] = ...
                orthb(reshape(v.u(:,:,end-nb+1:end),[],nb),reshape(v.u(:,:,1:end-nb),[],k-nb));
        elseif range == 0 % Do full orthogonalization (u-vector and L-part)
            % TODO: Make more efficient?
            % Very expensive!
            LL = zeros(size(C{1},1),0);
            for it = 1:d_C
                LL = [LL,C{it}*v.Q];
            end
            r = size(v.Q,2);
            B = blkdiag(speye((d_phi+1)*r),kron(speye(d_psi),LL'*LL));
            [u_,h,normm] = ...
                orthb(reshape(v.u(:,:,end-nb+1:end),[],nb),reshape(v.u(:,:,1:end-nb),[],k-nb),[],[],[], ...
                B);
        else
            % Orhtogonalize only part of the range of the vector
            sel = range>d_phi+1;
            range_ = [range(~sel)  ...
                kron(range(sel),ones(1,d_C)) + ...
                kron(ones(1,sum(sel)),0:d_C-1) + ...
                kron(0:d_C-1:(sum(sel)-1)*(d_C-1),ones(1,d_C))];
            % Orthogonalize the vectors
            [u_,h,normm] = ...
                orthb(reshape(v.u(:,:,end-nb+1:end),[],nb), ...
                reshape(v.u(:,:,1:end-nb),[],k-nb), ...
                [],reshape(v.u(:,range_,end-nb+1:end),[],nb), ...
                reshape(v.u(:,range_,1:end-nb),[],k-nb));
        end
        % Reshape into the good form
        v.u(:,:,end-nb+1:end) = reshape(u_,r,d,nb);
    end

    function v = bs(B,s)
        % Operator (AA-s*BB)\bb, for the case where we have a vector of the form
        % bb = [B;0;0;...;0];
        
        % Get current function values;
        if L.ist == 2
            fi = conj(f(s));
        else
            fi = f(s);
        end
        
        d = d_phi+1+d_psi*d_C;
        v.Q = [];
        v.u = zeros(0,d,0);
        A = @(s) [speye(size(AA_L,1),1) AA_L(:,d_C+1:end)-s*BB_L(:,d_C+1:end)];
        % Solve large scale system of the first block, form Q and u
        for it = 1:size(B,2)
            x = backslash(@eval_mat,B(:,1),fi);
            [q,h0,h1] = orthb(x,v.Q);
            if isempty(h0) || norm(h0) > 1e-15
                v.Q = [v.Q,q];
                v.u = reshape([reshape(v.u,it-1,[]);zeros(1,d*(it-1))],it,d,it-1);
                v.u(:,:,end+1) = backslash(A,[zeros(it-1,d);[h1 zeros(1,d-1)]],s,[],1);
            end
        end
    end

    function [v,lud] = bsB(v,s,nb,lud)
        % Operator (AA-s*BB)\BB*b in low-rank format
        % Handle the LU information, by adding the low-rank information
        if nargin < 4 || ~isfield(lud,'low'), lud.low.stat = 1; end
        % Get current function values;
        if L.ist == 2
            fi = conj(f(s));
        else 
            fi = f(s);
        end
        % Solve low-rank system for the rhs first block:
        % H_L = [P_L 0 ... 0] (AA-sBB)^{-1}, by replacing by
        % H_L = ([P_L+psi_0 0 ... 0]-AA_L,0-sBB_L,0)
        % [[1 0 ...0];[AA_L,1-sBB_L,1]]^{-1}, with AA_L,0 the first block
        % row, and AA_L,1 the rest. Note that P_L(s) = fi in this
        % formulation
        A = @(s) [speye(size(AA_L,1),1) AA_L(:,d_C+1:end)-s*BB_L(:,d_C+1:end)];
        
        b = kron(fi,speye(size(AA_L,1),1))-(AA_L(:,1:d_C)-s*BB_L(:,1:d_C));
        [H_L,lud.low] = backslash(A,b,s,lud.low);
        % TODO: add all of the blocks at once..
        for itt = 1:nb % Loop over block vectors
            % Construct the right hand side
            u_ = v.u(:,:,end-nb+itt)*(BB_L*[speye(d_C,d_C);H_L(2:end,:)]);
            b = 0;
            for it = 1:d_C
                qq = v.Q*u_(:,it);
                b = b + C{it}*qq;
            end
            % Solve large scale system of the first block, add to Q and expand u
            [x,lud] = backslash(@eval_mat,b,fi,lud);
            [r,d,k] = size(v.u);
            [q,h0,h1] = orthb(x,v.Q);
            
            if norm(h0) > 1e-15 % Check to see if x was already in Q
                % No, so add the new vector to the basis
                v.Q = [v.Q,q];
                v.u = reshape([reshape(v.u,r,[]);zeros(1,d*k)],r+1,d,k);
                % Solve the low-rank system for the rest of the vectors
                v.u(:,:,end+1) = backslash([],[[h0;h1] v.u(:,:,end-itt+1)*BB_L(:,d_C+1:end)],[],lud.low,1);
            else
                % Solve the low-rank system for the rest of the vectors
                v.u(:,:,end+1) = backslash([],[v.Q'*x v.u(:,:,end-itt+1)*BB_L(:,d_C+1:end)],[],lud.low,1);
            end
        end
    end

    function [AA,BB,vv] = large(v)
        % Get the large pencil matrix AA, BB explicitely. Warning: might be
        % expensive, for testing purpouses only
        
        Tcl = kron(speye(d_psi),speye(d_C,1)).';
        Tcr = kron(speye(d_psi+1),speye(d_C,1));
        
        M_psi = (Tcl*AA_L(d_phi+2:end,1:(d_psi+1)*d_C)*Tcr);
        N_psi = (Tcl*BB_L(d_phi+2:end,1:(d_psi+1)*d_C)*Tcr);
        
        AAA_L = AA_L.';
        BBB_L = BB_L.';
        n = size(C{1},1);
        
        Top = sparse(n*(d_psi+1),n*(d_phi+1));
        for it = 1:d_C
            rows = it:d_C:(d_psi+1)*d_C;
            Top = Top + kron(AAA_L(rows,1:d_phi+1),C{it});
        end
        AA = [[Top;kron(AAA_L(d_C*(d_psi+1)+1:end,1:d_phi+1),speye(n,n))],[kron(M_psi.',speye(n,n));sparse(d_phi*n,d_psi*n)]];
        
        Top = sparse(n*(d_psi+1),n*(d_phi+1));
        for it = 1:d_C
            rows = it:d_C:(d_psi+1)*d_C;
            Top = Top + kron(BBB_L(rows,1:d_phi+1),C{it});
        end
        BB = [[Top;kron(BBB_L(d_C*(d_psi+1)+1:end,1:d_phi+1),speye(n,n))],[kron(N_psi.',speye(n,n));sparse(d_phi*n,d_psi*n)]];
        
        if nargin == 1
            vv = zeros(((d_phi+1)+d_psi)*n,size(v.u,3));
            r = size(v.u,1);
            vv(1:n*(d_phi+1),:) = reshape(v.Q*reshape(v.u(:,1:d_phi+1,:),r,[]),[],size(v.u,3));
            for it = 1:d_psi
                for it_in = 1:d_C
                    vv(n*(d_phi+1)+((it-1)*n)+1:n*(d_phi+1)+it*n,:) = ...
                        vv(n*(d_phi+1)+((it-1)*n)+1:n*(d_phi+1)+it*n,:) + ...
                        reshape(C{it_in}*v.Q*reshape(v.u(:,d_phi+1+(it-1)*d_C+it_in,:),r,[]),[],size(v.u,3));
                end
            end
        end
    end

    function ip = ipp(v,w)
        kv = size(v.u,3);
        dv = size(v.u,2);
        rv = size(v.u,1);
        kw = size(w.u,3);
        
        C_ = w.Q'*v.Q;
        vuc = reshape(C_*reshape(v.u,rv,dv*kv),size(C_,1),dv,kv);
        ip = reshape(w.u,[],kw)'*reshape(vuc,[],kv);
    end

    function [A_,B_,vu_A] = project(v,w,trsp)
        % Project the linear pencil as w'(AA-sBB)v. If only one input is
        % given, the CORK pencil is presumed with V given
        % TODO: generalize this last case
        
        % If the pencil is in transpose mode, calculate
        % conj(v)'(AA^T-sBB^T) conj(w)
        if nargin == 1
            w = v;
        end
        if L.ist == 2
            v_ = v;
            v.Q = w.Q;
            v.u = w.u;
            w.Q = v_.Q;
            w.u = v_.u;
            clear v_;
        elseif L.ist == 1
            error('Cannot use project on transposed pencil. Use L.transpose firts, or conjugate transpose instead.')
        end
        
        kv = size(v.u,3);
        rv = size(v.u,1);
        kw = size(w.u,3);
        [rr,dd,kk] = size(v.u);
        vu_A = permute(reshape(reshape(permute(v.u,[1,3,2]),rr*kk,dd)*AA_L,rr,kk,size(AA_L,2)),[1,3,2]);
        vu_B = permute(reshape(reshape(permute(v.u,[1,3,2]),rr*kk,dd)*BB_L,rr,kk,size(BB_L,2)),[1,3,2]);
        
        A_ = zeros(kw,kv);
        B_ = zeros(kw,kv);
        for it = 1:d_C
            sel_v = it+(0:1:d_psi)*d_C;
            sel_w = 1:d_psi+1;
            if trsp == 2
                C_ = w.Q'*C{it}*v.Q;
                A_ = A_ + reshape(w.u(:,sel_w,:),[],kw)'*reshape(C_*reshape(vu_A(:,sel_v,:),rv,[]),[],kv);
                B_ = B_ + reshape(w.u(:,sel_w,:),[],kw)'*reshape(C_*reshape(vu_B(:,sel_v,:),rv,[]),[],kv);
                 if d_psi > 0
                    sel_v = (d_psi+1)*d_C+1:(d_psi+1)*d_C+d_phi;
                    sel_w = d_psi+1+it+(0:1:d_phi-1)*d_C;
                    A_ = A_ + reshape(w.u(:,sel_w,:),[],kw)'*reshape(C_*reshape(vu_A(:,sel_v,:),rv,[]),[],kv);
                    B_ = B_ + reshape(w.u(:,sel_w,:),[],kw)'*reshape(C_*reshape(vu_B(:,sel_v,:),rv,[]),[],kv);
                 end
            elseif trsp == 1
                C_ = w.Q.'*C{it}*v.Q;
                A_ = A_ + reshape(w.u(:,sel_w,:),[],kw).'*reshape(C_*reshape(vu_A(:,sel_v,:),rv,[]),[],kv);
                B_ = B_ + reshape(w.u(:,sel_w,:),[],kw).'*reshape(C_*reshape(vu_B(:,sel_v,:),rv,[]),[],kv);
                if d_psi > 0
                    sel_v = (d_psi+1)*d_C+1:(d_psi+1)*d_C+d_phi;
                    sel_w = d_psi+1+it+(0:1:d_phi-1)*d_C;
                    A_ = A_ + reshape(w.u(:,sel_w,:),[],kw).'*reshape(C_*reshape(vu_A(:,sel_v,:),rv,[]),[],kv);
                    B_ = B_ + reshape(w.u(:,sel_w,:),[],kw).'*reshape(C_*reshape(vu_B(:,sel_v,:),rv,[]),[],kv);
                end
            end
        end
        if d_psi == 0
            sel_v = (d_psi+1)*d_C+1:(d_psi+1)*d_C+d_phi;
            sel_w = 2:size(v.u,2);
            if trsp == 2
                A_ = A_ + reshape(w.u(:,sel_w,:),[],kw)'*reshape(vu_A(:,sel_v,:),[],kv);
                B_ = B_ + reshape(w.u(:,sel_w,:),[],kw)'*reshape(vu_B(:,sel_v,:),[],kv);
            elseif trsp == 1
                A_ = A_ + reshape(w.u(:,sel_w,:),[],kw).'*reshape(vu_A(:,sel_v,:),[],kv);
                B_ = B_ + reshape(w.u(:,sel_w,:),[],kw).'*reshape(vu_B(:,sel_v,:),[],kv);
            end
        end
        if L.ist == 2
            A_ = A_';
            B_ = B_';
        end
    end

    function [AA_L,BB_L,d_psi,d_phi,d_C,A_L,B_L] = initialize(p)
        % Construct AA_L and BB_L (in transposed form!), in rank-independent form, based on structure p
        % p contains all the information on the polynomial:
        % p.M_psi, p.N_psi, p.M_phi, p.N_phi (Linear relation matrices)
        % p.TA, p.TB (Transformations from C-matrices to A,B-matrices)
        % p.L_phi, p.L_psi (Position of A,B-matrices in pencil)
        
        % Inititialize dimensions
        d_C = length(C);
        d_phi = size(p.M_phi,1);
        d_psi = size(p.M_psi,1);
        
        % Forming of A_L,B_L by placing p.TA and p.TB on correct positions (using p.L_phi,p.L_psi)
        A_L = sparse(kron(ones(1,length(p.L_psi)),(1:d_C)') + kron((p.L_psi-1)*d_C,ones(d_C,1)),kron(p.L_phi,ones(d_C,1)),p.TA);
        B_L = sparse(kron(ones(1,length(p.L_psi)),(1:d_C)') + kron((p.L_psi-1)*d_C,ones(d_C,1)),kron(p.L_phi,ones(d_C,1)),p.TB);
        
        % Forming AA_L, BB_L
        AA_L = [[A_L.';kron(p.M_psi,speye(d_C))] [p.M_phi.';sparse((d_psi)*d_C,(d_phi))]];
        BB_L = [[B_L.';kron(p.N_psi,speye(d_C))] [p.N_phi.';sparse((d_psi)*d_C,(d_phi))]];
    end

    function A = eval_mat(f)
        % Evaluate the non-linear matrix function, given the evaluated
        % non-linear matrix factors
        A = sparse(1,1);
        for it = 1:length(f)
            A = A + f(it)*C{it};
        end
    end
end