function [ v,h,norm_new,v_ ] = orthb( v,V,reorth,v_,V_,B )
%ORTHB - Orhtonormalize vector(s) v (induvidually) against V. If V = [],
%        each of the vectors are just normalized.
%
% Inputs:
%   v  - Vector(s) to be orthonormalized. Use: vectors are not
%   orthonormalized against each other, but orthogonalized/normalized
%   induvidually, oposite basis V if given!
%   V  - If given, orhtonormal matrix containing vectors that span a basis to be
%   orthonormalized against
%   reorth - Vector containing reorthogonalization constants
%   v_ - If given, this vector is used to compute the inner products 
%   instead of the full vector v.
%   V_ - As v_, if given, it is the partial V to be used for the inner
%   product
%   B - Alternatively, a B matrix can be given, so that the alternative
%   B-inner product v'*B*V is used. In this case, v_ and V_ inputs are ignored.
%
% Note, (v_,V_)-combo or B-innerproduct not possible with multiple vectors
% in v. (TODO!)
%
% Outputs:
%   v - Orthonormalized vector
%   h - Normalization constants so that v = V*h
%   norm_new - Norm of v before normalization
%   v_ - Normalized alternative vector (if given as input)

% Case of no basis given is just normalization
if isempty(V) && nargin < 4
    norm_new = zeros(1,size(v,2));
    for it = 1:size(v,2)
        norm_new(it) = norm(v(:,it));
    end
    %v = bsxfun(@rdivide,v,norm_new);
    v = v./norm_new;
    h = [];
    return;
elseif isempty(V) && nargin == 6 % When B is given, use the B-inner product
    norm_new = sqrt(v'*B*v);
    v = v/norm_new;
    h = [];
    return;
elseif isempty(V) % When v_ is given, use the norm of v_ instead
    norm_new = norm(v_);
    v = v/norm_new;
    h = [];
    return;
end
% Set reorthogonalization vector, if not given
if nargin < 3 || isempty(reorth)
    reorth = 1/sqrt(2)*ones(1,3);
end
% Number of reorthogonalizations
nreorth = length(reorth);
if nargin < 4 % Case of no alternative inner product (Gramm-Smidth)
    h = V'*v;
    norm_old = norm(v);
    v = v - V*h;
    norm_new = norm(v);
    it = 0;
    while (it < nreorth) && (norm_new < reorth(it+1)*norm_old)
        h_new = V'*v;
        v = v - V*h_new;
        h = h + h_new;
        norm_old = norm_new;
        norm_new = norm(v);
        it = it+1;
    end
    [v,norm_new] = qr(v,0);
elseif nargin == 6 % Use the B-innerproduct
    h = V'*(B*v);
    norm_old = sqrt(v'*(B*v));
    v = v - V*h;
    norm_new = sqrt(v'*(B*v));
    it = 0;
    while (it < nreorth) && (norm_new < reorth(it+1)*norm_old)
        h_new = V'*(B*v);
        v = v - V*h_new;
        h = h + h_new;
        norm_old = norm_new;
        norm_new = sqrt(v'*(B*v));
        it = it+1;
    end
    v = v/norm_new;
else % Use the v_ and V_
    h = V_'*v_;
    norm_old = norm(v_);
    v_ = v_ - V_*h;
    v = v - V*h;
    norm_new = norm(v_);
    it = 0;
    while (it < nreorth) && (norm_new < reorth(it+1)*norm_old)
        h_new = V_'*v_;
        v_ = v_ - V_*h_new;
        v = v - V*h_new;
        h = h + h_new;
        norm_old = norm_new;
        norm_new = norm(v_);
        it = it+1;
    end
    v = v/norm_new;
    v_ = v_/norm_new;
end
end
