function [ V,info ] = rkrylov( A,v,opts )
%Rkrylov - Rational Krylov iterations
%
% V = krylov( A,v ) - Krylov iterations that result in Krylov
% basis V. Krylov iterations performed using operator A, starting
% 'vectorelement' v.
%
% A is a function handle A(v,s), or A(v,s,LU) that accepts v, shifts s
%    and maybe LU-data. In the case of eigenvalue computation and automatic
%    residue computation, A should be the implementation of an operator (AA-sBB)\(BB*v)
%    (i.e. krylov is the rational Krylov method for solving the generalized eigenvalue problem (AA-s*BB)).
%
% V = rkrylov( A,v,opts) - wit the following valid options:
%       opts.m         - Number of iterations [10]
%       opts.k         - Number of eigenvalues/eigenvectors calculated [2]
%       opts.sig       - Shifts [0]
%       opts.tol       - Relative tolerance for eigenvalues [1e-14]
%       opts.nb        - Number of blocks in case of block Krylov method [1]
%       opts.get_res   - Get the residues in each iteration
%       opts.resfun    - Function for calculating residual. If no residual function is given,
%                        it is estimated automatically [-].
%       opts.orth_fun  - Orthogonalization function. If no function is given orthb is used. [-]

%       opts.concafun  - Function for concatenating two elemens. If no
%                        function is given cat(2,v,V) is used. Note that
%                        opts.concafun(v,[]) should be equivalent to
%                        stack(v) (i.e. increase dimension by 1) [-]
%       opts.getres    - Boolean indicating whether one computes the residual
%                        in each step [false]
%       opts.verbose   - Output the iteration count and the residuals [false]
%
% [lambda,V,info] = ARNOLDI( A,v,opts ) - 'V' contains the basis vectors build during
% the iterations. 'info' contains data from the iterations:
%       info.Lambda - The eigenvalue estimates in each iteration
%       info.res    - Residuals from all the iterations
%       info.times  - Timings per iteration

% Check options and set parameters
if nargin < 3, opts = struct; end
[ k,m,nb,sig,orth_fun,getres,resfun,conca_fun,mult_fun,subr_fun,restarts,filter,verbose ] = checkOpts( opts,v );
% Initialzie all the variables
[ A,V,v,sig,sig_,H,K,lud_vec,resfun,do_rk,info ] = initialize( A,v,m,sig,orth_fun,resfun,conca_fun,getres );
% Begin iterations
it = 1;
for itt = 1:m-1
    t_start = tic;
    % Apply the operator
    [v,lud_vec{sig_(it)==sig}] = A(v,sig_(it),lud_vec{sig_(it)==sig});
    % Orthonormalize
    cols = (it-1)*nb+1:it*nb;
    rows = it*nb+1:it*nb+nb;
    if getres
        [ v,H(1:it*nb,cols),H(rows,cols) ] = orth_fun( v,V );
    else
        [ v ] = orth_fun( v,V );
    end
    % Concatenate
    V = conca_fun( v,V );
    
    % Get residue if needed
    if restarts(it)
        K(1:it*nb,cols) = H(1:it*nb,cols)*sig_(it) + rot90(eye(it*nb,nb),2);
        K(it*nb+1:it*nb+nb,cols) = H(it*nb+1:it*nb+nb,cols)*sig_(it);
        [TT,SS,YY,ZZ,S,~] = qz(K(1:it,1:it),H(1:it,1:it));
        lambda = ordeig(TT,SS);
        res = resfun(lambda,H,K,S,V,k,it);
        m_want = 40;
        
        selres = zeros(it,1);
        [~,ind] = sort(res);
        selres(ind) = it:-1:1;
        
        [~,idx] = sort(abs(lambda-mean(sig)));
        sell = zeros(it,1);
        sell(idx(1:m_want)) = selres(idx(1:m_want));
        
        [TT,SS,YY,ZZ] = ordqz(TT,SS,YY,ZZ,sell);
        m_con = sum(res(idx(1:m_want))<1e-12);
        
        hh = H(it+1,1:it)*ZZ;
        kk = K(it+1,1:it)*ZZ;
        
        H(1:m_want+1,1:m_want) = [SS(1:m_want,1:m_want);[zeros(1,m_con),hh(m_con+1:m_want)]];
        K(1:m_want+1,1:m_want) = [TT(1:m_want,1:m_want);[zeros(1,m_con),kk(m_con+1:m_want)]];
        
        v = mult_fun(v,YY(1:m_want,:)');
        
        it = m_want;
    elseif filter(itt)
        [QQ,RR] = qr(H(1:it+1,1:it),0);
        V = mult_fun(V,QQ);
        v = subr_fun(V);
        H(1:it,1:it-1) = RR*QQ(1:end-1,1:end-1);
        K(1:it,1:it-1) = H(1:it,1:it-1)*sig_(it) + rot90(eye(it,1),2);
        it = it-1;
        
        if getres
            if do_rk
                [G,F,~,~,S,~] = qz(K(1:it*nb,1:it*nb),H(1:it*nb,1:it*nb));
                lambda = ordeig(G,F);
            else
                [S,lambda] = eig(H(1:it*nb,1:it*nb));
                lambda = diag(lambda);
            end
            res = resfun(lambda,H,K,S,V,k,it*nb);
            info.Res(1:it,it) = res(1:it);
            info.Lambda(1:it,it) = lambda(1:it);
        end
    elseif getres
        if do_rk
            K(1:it*nb,cols) = H(1:it*nb,cols)*sig_(it) + rot90(eye(it*nb,nb),2);
            K(it*nb+1:it*nb+nb,cols) = H(it*nb+1:it*nb+nb,cols)*sig_(it);
            [G,F,~,~,S,~] = qz(K(1:it*nb,1:it*nb),H(1:it*nb,1:it*nb));
            lambda = ordeig(G,F);
        else
            [S,lambda] = eig(H(1:it*nb,1:it*nb));
            lambda = diag(lambda);
        end
        res = resfun(lambda,H,K,S,V,k,it*nb);
        info.Res(1:it,it) = res(1:it);
        info.Lambda(1:it,it) = lambda(1:it);
    end
    % Increase the running iteration index
    it = it+1;
    % Get the time
    info.times(it) = toc(t_start);
    % Print outputs
    if verbose && getres
        fprintf('It: %i, number of res < tol: %i \n',it,sum(res < 1e-13));
    elseif verbose
        fprintf('It: %i \n',it);
    end
end
% Set rest of info fields if get res was one
if getres
    [~,idx] = sort(res);
    info.lambda = lambda(idx(1:k));
    info.H = H(1:it,1:it-1);
    info.K = K(1:it,1:it-1);
end

    function [ k,m,nb,sig,orth_fun,getres,resfun,conca_fun,mult_fun,subr_fun,restarts,filter,verbose ] = checkOpts( opts,v )
        % Check the options and set values
        
        % Number of eigenvalues
        if isfield(opts,'k'), k = opts.k; else, k = 2; end
        % Number of iterations
        if isfield(opts,'m'), m = opts.m; else, m = 5*k; end
        if m < k, m=k; end
        % Shifts
        if isfield(opts,'sig'), sig = opts.sig;
        elseif ~isfield(opts,'sig'), sig = 0;
        end
        % Number of blocks
        if isfield(opts,'nb')
            nb = opts.nb;
        elseif ismatrix(v)
            nb = size(v,2);
        else
            nb = 1;
        end
        % Verbose
        if isfield(opts,'verbose')
            verbose = opts.verbose;
        else
            verbose = false;
        end
        % Orthogonalization function
        if isfield(opts,'orth_fun')
            orth_fun = opts.orth_fun;
        else
            orth_fun = @orthb;
        end
        % Get the residual
        if isfield(opts,'getres')
            getres = opts.getres;
        else
            getres = false;
        end
        % Residue function
        if isfield(opts,'resfun')
            resfun = opts.resfun;
        else, resfun = [];
        end
        % Vector specific functions
        if isfield(opts,'conca_fun')
            conca_fun = opts.conca_fun;
        else
            conca_fun = @(v,V) cat(2,V,v);
        end
        % Multiplication function
        if isfield(opts,'mult_fun')
            mult_fun = opts.mult_fun;
        else
            mult_fun = @(V,M) V*M;
        end
        % Restarts
        restarts=zeros(1,m);
        if isfield(opts,'restarts')
            restarts(opts.restarts:opts.restarts:end) = 1;
        end
        % Set implicit filter
        filter = zeros(1,m);
        if isfield(opts,'filter')
            filter(opts.filter.start:opts.filter.step:end) = 1;
        end
        
        if isfield(opts,'subr_fun')
            subr_fun = opts.subr_fun;
        else
            subr_fun = @(V) V(:,end);
        end
    end

    function [ A,V,v,sig,sig_,H,K,lud_vec,resfun,do_rk,info ] = initialize( A,v,m,sig,orth_fun,resfun,conca_fun,getres )
        % Initialize iteration vectors
        v = orth_fun(v,[]);  % Orthogonalization against no basis is equivalent to normalization
        
        % Get V as the concatenation of v with empty matrix (important for
        % v-elements that need to be stacked, i.e. in this case we presume
        % conca_fun(v,[]) is equivalent to stack(v)).
        V = conca_fun(v,[]);
        
        % Initialize H and K
        H = zeros(m*nb,m*nb-nb);
        K = zeros(m*nb,m*nb-nb);
        
        % Get unique shifts
        sig = unique(sig);
        % Get interpolation points for each m
        if size(sig,2) == 1, sig = sig.'; end
        if ~isempty(sig)
            % Repeat the interpolation points (simple, can be extended based on residues...)
            sig_ = repmat(sig,1,floor(m/length(sig)));
            if length(sig_) ~= m
                sig_ = [sig_,sig(1:m-length(sig_))];
            end
        else
            sig_ = 0;
        end
        
        % Initialze the LUdata
        lud.stat = 1;
        lud_vec = cell(1,length(sig));
        for it_ = 1:length(sig)
            lud_vec{it_} = lud;
        end
        
        % Reinitialize the operator
        if isnumeric(A)
            A = @(x,s,lud) deal(A*x,lud);
            do_rk = false;
        elseif nargin(A) == 1
            A = @(x,s,lud) deal(A(x),lud);
            do_rk = false;
        elseif nargin(A) == 2
            A = @(x,s,lud) deal(A(x,s),lud);
            do_rk = false;
        else
            do_rk = true;
        end
        
        % Reinitialize resfun
        if isempty(resfun) && do_rk
            resfun = @(lambda,H,K,S,V,k,it) abs((K(it+1,it)-lambda(1:it)*H(it+1,it)).*S(it,1:it).');
        elseif isempty(resfun)
            resfun = @(lambda,H,K,S,V,k,it) abs(lambda(1:it)*H(it+1,it).*S(it,1:it).');
        else
            resfun = @(lambda,H,K,S,V,k,it) resfun(lambda,V,H(1:it+nb,1:it)*S);
        end
        
        % Initialize time and residual information
        info.times = zeros(1,m-1);
        if getres
            info.Res = zeros(m-1,m-1);
            info.Lambda = zeros(m-1,m-1);
        end
    end
end
